/* ==== iwr-shade.c ==== */
/* Part of IWR: Imperfect World Renderer, by Tom Barnes-Lawrence */

#include "SDL.h" /* Need to compile with -I/usr/include/SDL or similar */
#include "iwr-shade.h"
#include "iwr-texture.h"

#define HF_FILE "iwr-hf.bmp"
/* Should eventually change HF to whatever format Palito uses */
#define IMG_FILE "iwr-img.bmp"

int main () {
    /* command-line arguments could include the filename for scheme
    code to describe the 3d texturing function, alternative
    unit-vector choices for the primary light-source (there's only
    going to be one), ambient-light levels, and so on.
    Initial version is going to take no arguments, have no texturing,
    and is going to load input from a specific bmp heightfield and
    save it to a specific bmp target.
    Later versions will take heightfield input from stdin in the
    same format Palito takes and put the output image to stdout
    in whatever format Palito expects (which is currently BMP too).*/
    SDL_Surface *hf_surface; /* The heightfield to render */
    SDL_Surface *img_surface; /* The rendered (shaded) image */
    if (SDL_Init(SDL_INIT_VIDEO) ) {
      fprintf(stderr,"Couldn't init SDL! Exiting.");
      exit(1);
    }
    hf_surface=SDL_LoadBMP(HF_FILE);
    if (hf_surface==NULL) {
      fprintf(stderr,"Couldn't load heightfield file!! Exiting.");
      exit(1);
    }
    img_surface=iwr_shade(hf_surface);
    if(SDL_SaveBMP(img_surface,IMG_FILE)) {
      fprintf(stderr,"Couldn't save the rendered file!! Exiting.");
      exit(1);
    }
    SDL_Quit();
}


SDL_Surface *iwr_shade(SDL_Surface *heightfield)
{
    SDL_Surface *new_surface;
    Uint32 rmask, gmask, bmask, amask, hmask; /* but I don't want alpha! */
    Uint32 temp;
    double illum; /* Illumination level, in range 0-1 */
    Uint8 texr,texg,texb; /* colours found from texture */
    Uint16 colr,colg,colb; /* final colours to put into image */
    Uint16 hpitch,ipitch; /* Pitch of heightfield + image */
    Uint8 hbpp,ibpp; /* *bytes* per pixel */
    Uint8 hloss,hshift; /* For accessing heightfield */
    Uint8 rloss,gloss,bloss,rshift,gshift,bshift; /* For writing image */
    void *hlinestart,*ilinestart; /* Pointers to starts of lines */
    void *hscan, *iscan; /* Pointers to pixels of surface */
    int iwidth,iheight; /* Dimensions of image, 1 less than heightfield */
    int x,y; /* Coords of point being scanned */
    int a,b,c; /* Heights of points in triangle being examined */
    double zb,zc; /* height deltas of triangle being examined */
    double xa,ya,za; /* the surface normal */
    double xl=-0.4,yl=-0.1,zl=0.8; /* vector pointing to light source */
    double lena,lenl;
    lenl=sqrt((xl*xl)+(yl*yl)+(zl*zl));
    xl/=lenl;yl/=lenl;zl/=lenl; /* scanle light vector to unit vector */

    iwidth=(heightfield->w)-1;
    iheight=(heightfield->h)-1;

    /* Get mask info...? */
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
    rmask = 0xff000000;
    gmask = 0x00ff0000;
    bmask = 0x0000ff00;
    amask = 0x000000ff;
#else
    rmask = 0x000000ff;
    gmask = 0x0000ff00;
    bmask = 0x00ff0000;
    amask = 0xff000000;
#endif
    /* Allocate the surface */
    new_surface=SDL_CreateRGBSurface(SDL_SWSURFACE,iwidth,iheight,32,rmask,gmask,bmask,amask);
    if(new_surface==NULL) {
      fprintf(stderr,"Couldn't produce new SDL surface!! Exiting.");
      exit(1);
    }
    SDL_LockSurface(heightfield);
    SDL_LockSurface(new_surface);
    hpitch=heightfield->pitch;  ipitch=new_surface->pitch;
    hbpp=heightfield->format->BytesPerPixel;
    if (hbpp==1) {
      fprintf(stderr,"Heightfield should be >8bpp!! Exiting.");
      exit(1);
    }
    ibpp=new_surface->format->BytesPerPixel;
    hmask=heightfield->format->Rmask;
    hloss=heightfield->format->Rloss;
    hshift=heightfield->format->Rshift;
    rmask=new_surface->format->Rmask; gmask=new_surface->format->Gmask;
    bmask=new_surface->format->Bmask;
    rloss=new_surface->format->Rloss; rshift=new_surface->format->Rshift;
    gloss=new_surface->format->Gloss; gshift=new_surface->format->Gshift;
    bloss=new_surface->format->Bloss; bshift=new_surface->format->Bshift;
    /* Scan through all the points to shade */
    hlinestart=heightfield->pixels;
    ilinestart=new_surface->pixels;
    for (y=0;y<iheight;y++)
    {
       /* hlinestart=(y*hpitch)+heightfield->pixels; */
       /* ilinestart=(y*ipitch)+new_surface->pixels; */
       hscan=hlinestart;iscan=ilinestart;
       for (x=0;x<iwidth;x++,hscan+=hbpp,iscan+=ibpp)
       {
          /* get surface normal of triangle using hscan, hscan+hbpp,
          and hscan+hpitch.
          If HF is 1 byte per pixel, use index as height. Else use
          Red component as height (ignore other channels).*/
          temp=*((Uint32 *)hscan)&hmask;
          temp=temp>>hshift; /* I'm sure the SDL docs are wrong...*/
          temp=temp<<hloss;
          a=(Uint8 )temp;
          temp=*((Uint32 *)hscan+hbpp)&hmask;
          temp=temp>>hshift;temp=temp<<hloss;
          b=(Uint8 )temp;
          temp=*((Uint32 *)hscan+hpitch)&hmask;
          temp=temp>>hshift;temp=temp<<hloss;
          c=(Uint8 )temp;
          /* Now have heights of 3 corners. Work out deltas around "a"
          to get cross-product (giving surface normal) */

          /* 3 x 1 = x2,y2,z2
            x2= y3*z1 - z3*y1
            y2= z3*x1 - x3*z1
            z2= x3*y1 - y3*x1

            2 is normal at a;  3 is vector to b  1 is vector to c
          so rename x2,y2,z2= xa,ya,za,
          x1,y1,z1= xc,yc,zc= 0,1,c-a
          x3,y3,z3= xb,yb,zb= 1,0,b-a */
          zb=(b-a)/4;zc=(c-a)/4; /* Should mebbe scale down, by 1/3 or something */
          xa=(0*zc)-(zb*1); /*-zb */
          ya=(zb*0)-(1*zc); /*-zc */
          za=(1*1)-(0*0);   /* 1 */
          /* Hmm, that's a bit more simple than a generic cross-product...*/

          /* get dot-product of surface normal and light direction */
          /* illum=( (xa,ya,za) DOT (xl,yl,zl)); */
          /* the surface normal and light direction must be *unit* vectors*/
          lena=sqrt((xa*xa)+(ya*ya)+(za*za)); /* get length */
          xa/=lena;ya/=lena;za/=lena; /* scale to unit vector */
          illum=(xa*xl)+(ya*yl)+(za*zl); /* illumination level,0-1 */
          /* Nope, turns out it can be negative too...*/
          illum=(illum<0)?0:illum;
          illum*=0.9;illum+=0.1;
          /* get texture of point at (x,y,hscan) */
          iwr_getrgbtexture(x,y,a, &texr,&texg,&texb);/* Find texture at point*/
          /* work out lighting from calculated dot-product */
          /* Apply lighting to texture */
            /* Initially don't do specularity.
            Assume full illumination gives the texture colours, whilst
            none gives black. In between gives interpolated level. */
          /* colr=(texr*illum)/255; */
          colr=iwr_unblacken(texr,illum);
          colg=iwr_unblacken(texg,illum);
          colb=iwr_unblacken(texb,illum);
            /* When doing specularity, do interpolation between full-white
            and texture colour. This is most likely wrong. */
          /* Put the calculated colour into new_surface */
          temp=colr>>rloss;temp=temp<<rshift;*(Uint32 *)iscan|=(temp&rmask);
          temp=colg>>gloss;temp=temp<<gshift;*(Uint32 *)iscan|=(temp&gmask);
          temp=colb>>bloss;temp=temp<<bshift;*(Uint32 *)iscan|=(temp&bmask);
          *(Uint32*)iscan|=(0xFFFFFFFF&amask);
       }
       hlinestart+=hpitch;
       ilinestart+=ipitch;
    }
    SDL_UnlockSurface(heightfield);
    SDL_UnlockSurface(new_surface);
    return(new_surface);
}

